import os
import re
import argparse
import tempfile
import subprocess
import urllib.parse

import urwid

from wikicurses import formats, settings
from wikicurses.wiki import Wiki, WikiError
from wikicurses.htmlparse import parseDisambig


def executeCommand(cmd):
    try:
        subprocess.call(cmd)
    except FileNotFoundError:
        ex.notify("Error: command '" + cmd[0] + "' not found.")
    loop.screen.clear() # Completely redraw screen after external command


def setTerminalWindowTitle(title):
    print('\33]0;' + title + ' - wikicurses\a')


def tabComplete(text, matches):
    if not matches:
        return text
    matches = sorted(matches, key=len)
    if matches[0] == text and len(matches) > 1:
        match = matches[1]
    else:
        match = matches[0]
    return match


class SearchBox(urwid.Edit):
    title = "Search"

    def keypress(self, size, key):
        if key == 'enter':
            closeOverlay()
            openPage(self.edit_text)
        elif key == 'tab':
            matches = wiki.search_sugestions(self.edit_text)
            match = tabComplete(self.edit_text, matches)
            self.edit_text = match
            self.edit_pos = len(match)
        elif key == 'esc':
            closeOverlay()
        else:
            return super().keypress(size, key)


class SelectorBox(urwid.ListBox):

    def __init__(self):
        def selectButton(radio_button, new_state, parameter):
            if new_state:
                closeOverlay()
                self._select(parameter)

        super().__init__(urwid.SimpleFocusListWalker([]))
        buttons = []
        for i, item in enumerate(self._items()):
            if isinstance(item, urwid.Widget):
                self.body.append(item)
                continue
            elif isinstance(item, tuple):
                name, selected, parameter = item
            else:
                parameter = name = item
                selected = False
            self.body.append(urwid.RadioButton(buttons, name, selected,
                                               selectButton, parameter))
            if selected:
                self.set_focus(i)

    def keypress(self, size, key):
        if key == 'esc':
            closeOverlay()
        else:
            return super().keypress(size, key)


class Toc(SelectorBox):
    title = "Table of Contents"

    def _items(self):
        for name, ind in mainwidget.body.widgetnames:
            yield name, mainwidget.body.body.focus >= ind, ind

    def _select(self, index):
        mainwidget.body.body.set_focus(index)


class Bmarks(SelectorBox):
    title = "Bookmarks"

    def _items(self):
        self.deleted = []
        return wiki.bmarks

    def _select(self, name):
        openPage(name)

    def keypress(self, size, key):
        # Undo Delete
        if key == 'u' and self.deleted:
            index, item = self.deleted.pop()
            wiki.bmarks.add(item.label)
            self.body.insert(index, item)
            self.set_focus(index)
        elif key in ('meta [', 'x') and self.focus:
            wiki.bmarks.discard(self.focus.label)
            self.deleted.append((self.focus_position, self.focus))
            self.body.remove(self.focus)
        else:
            return super().keypress(size, key)


class Links(SelectorBox):
    title = "Links"

    def _items(self):
        return page.links

    def _select(self, name):
        openPage(name)


class Iwlinks(SelectorBox):
    title = "Interwiki Links"

    def _items(self):
        netlocname = [(urllib.parse.urlparse(url).netloc, name)
                       for name, url in page.iwlinks]
        netlocs = set(netloc for netloc, name in netlocname)
        for netloc in netlocs:
            yield urwid.Text(netloc)
            # If the name in blank, the link refers to the site's Main Page
            yield from (j or "Main page" for i, j in netlocname if i == netloc)

    def _select(self, name):
        if name == "Main page":
            name = ''
        url = dict(page.iwlinks)[name]
        openWiki(Wiki.fromPageUrl(url))
        openPage(name)


class Langs(SelectorBox):
    title = "Languages"

    def _items(self):
        return page.langlinks.keys()

    def _select(self, lang):
        url, name = page.langlinks[lang]
        openWiki(Wiki.fromPageUrl(url))
        openPage(name)


class Extlinks(SelectorBox):
    title = "External Links"

    def _items(self):
        return page.extlinks

    def _select(sel, url):
        executeCommand([os.environ.get('BROWSER', 'lynx'), url])


class Wikis(SelectorBox):
    title = "Wikis"

    def _items(self):
        for name, url in settings.wikis().items():
            yield name, wiki.siteurl == url, name

    def _select(self, name):
        openWiki(name)
        openPage()


class Feeds(SelectorBox):
    title = "Feeds"

    def _items(self):
        return wiki.list_featured_feeds()

    def _select(self, feed):
        openPage(feed, True)


class Ex(urwid.Edit):
    mode = None
    highlighted = ''

    def highlightText(self, text):
        self.highlighted = text
        if text:
            mainwidget.body.search(text)
        else:
            mainwidget.body.unsearch()

    def keypress(self, size, key):
        if key == 'esc' or (key == 'backspace' and not self.edit_text):
            if self.mode == 'search':
                self.highlightText(self.previous_highlight)
            self.exitexmode()
        elif key == 'tab' and self.mode == 'ex':
            matches = [i for i in cmds if i.startswith(self.edit_text)]
            match = tabComplete(self.edit_text, matches)
            self.edit_text = match
            self.edit_pos = len(match)
        elif key == 'enter':
            if self.mode == 'ex':
                matches = [i for i in cmds if i.startswith(self.edit_text)]
                match = tabComplete(self.edit_text, matches)
                processCmd(*match.split())
            if self.mode == 'search':
                self.highlightText(self.edit_text)
            self.exitexmode()
        else:
            returnval = super().keypress(size, key)
            # Highlight after running super().keypress() so that edit_text is
            # up to date.
            if self.mode == 'search':
                self.highlightText(self.edit_text)
            return returnval

    def exitexmode(self):
        self.set_caption('')
        self.set_edit_text('')
        mainwidget.set_focus('body')
        self.mode = None

    def enterexmode(self):
        mainwidget.set_focus('footer')
        self.set_caption(':')
        self.mode = 'ex'

    def entersearchmode(self):
        self.previous_highlight = self.highlighted
        mainwidget.set_focus('footer')
        self.set_caption('/')
        self.mode = 'search'

    def notify(self, text):
        self.set_edit_text(text)


class StandardKeyBinds:

    def keypress(self, size, key):
        returnval = None
        maxcol, maxrow = size
        ex.notify('') # Clear any notification

        cmdmap = settings.conf['keymap']
        if key == ':':
            ex.enterexmode()
        if key == '/':
            # Disable search for Disambig/Results page
            if isinstance(mainwidget.body, Pager):
                ex.entersearchmode()
        elif key in ('g', 'home'):
            self.change_focus(size, 0, 0)
        elif key in ('G', 'end'):
            self.change_focus(size, len(self.body) - 1)
            offset = maxrow - self.focus.rows((maxcol,))
            self.change_focus(size, len(self.body) - 1, offset)
        elif key in cmdmap and cmdmap[key]:
            processCmd(cmdmap[key])
        else:
            returnval = super().keypress(size, key)

        # Set progress percentage
        lens = [i.rows((maxcol,)) for i in self.body]
        offset, inset = self.get_focus_offset_inset(size)
        # Number of the first line on the screen
        current_line = sum(lens[:self.body.focus]) - offset + inset
        position = current_line / (sum(lens) - maxrow) * 100
        progress.set_text(str(round(position)) + '%')

        return returnval

    def mouse_event(self, size, event, button, col, row, focus):
        if button == 4:
            self.keypress(size, 'up')
        if button == 5:
            self.keypress(size, 'down')
        else:
            return super().mouse_event(size, event, button, col, row, focus)
        return True


class Disambig(StandardKeyBinds, SelectorBox):
    widgetnames = []

    def __init__(self, html):
        self.sections = parseDisambig(html)
        super().__init__()

    def _items(self):
        for title, items in self.sections.items():
            if title:
                yield urwid.Text(['\n', ('h', title)], align='center')
            for name, text in items:
                yield (text, False, name) if name else urwid.Text(text)

    def _select(self, name):
        openPage(name)


class Results(StandardKeyBinds, SelectorBox):
    widgetnames = []

    def __init__(self, results):
        self.results = results
        super().__init__()

    def _items(self):
        return self.results

    def _select(self, title):
        openPage(title)


class Pager(StandardKeyBinds, urwid.ListBox):

    def __init__(self, page):
        super().__init__(urwid.SimpleFocusListWalker([]))
        self._content = page.content.copy()
        self._render()

    def _render(self):
        self.body.clear()
        self.widgetnames = [(page.title, 0)]
        curtext = []
        curh2 = ''
        prevalign = 'left'
        prevpadding = 0
        prevborder = False
        for tformat, text in self._content:
            align = settings.getColor(tformat, 'align', default='left')
            padding = settings.getColor(tformat, 'padding', default=0)
            border = settings.getColor(tformat, 'border', default=False)

            if (((align, padding, border) != (prevalign, prevpadding, prevborder)
                    and curtext) or (tformat & formats.h2 and not curh2)):
                # Also have new Text() for every h2 for TOC
                widget = urwid.Text(curtext, align=prevalign)
                if prevborder:
                    widget = urwid.LineBox(widget)
                if prevpadding:
                    widget = urwid.Padding(widget,
                        left=prevpadding, right=prevpadding)
                self.body.append(widget)
                curtext.clear()
            
            curtext.append((tformat, text))

            if tformat & formats.h2: 
                curh2 += text
            elif curh2:
                self.widgetnames.append((curh2, len(self.body) - 1))
                curh2 = ''

            prevalign = align
            prevpadding = padding
            prevborder = border
        if curtext:
            widget = urwid.Text(curtext, align=prevalign)
            if prevborder:
                widget = urwid.LineBox(widget)
            if prevpadding:
                widget = urwid.Padding(widget,
                    left=prevpadding, right=prevpadding)
            self.body.append(widget)

    def _add(self, text, attribute):
        if text:
            if self._content and self._content[-1][0] == attribute:
                self._content[-1][1] += text
            else:
                self._content.append([attribute, text])

    def search(self, findtext):
        self._content.clear()
        for attribute, text in page.content:
            cur = 0
            for match in re.finditer(findtext, text):
                start, end = match.start(), match.end()
                self._add(text[cur:start], attribute)
                self._add(text[start:end], attribute | formats.searchresult)
                cur = end
            if text[cur:]:
                self._add(text[cur:], attribute)
        self._render()

    def unsearch(self):
        self._content = page.content.copy()
        self._render()


def openPage(title=None, featured=False, browsinghistory=False):
    if not title:
        title = wiki.mainpage

    if not browsinghistory:
        global current
        if current < len(history)-1:
            del history[current+1:len(history)]
        history.append(title)
        current += 1

    global page
    if featured:
        page = wiki.get_featured_feed(title)
    else:
        page = wiki.search(title)
    # This is not as inefficient as it looks; Wiki caches results
    if not page.exists and wiki.search_sugestions(page.title):
        header.set_text('Results for ' + title)
        mainwidget.body = Results(wiki.search_sugestions(page.title))
    elif 'disambiguation' in page.properties:
        header.set_text(page.title + ': Disambiguation')
        mainwidget.body = Disambig(page.html)
    else:
        header.set_text(page.title)
        mainwidget.body = Pager(page)

    setTerminalWindowTitle(title)
    progress.set_text('0%')


def openWiki(name):
    global wiki
    if isinstance(name, Wiki):
        wiki = name
        return
    if not name:
        name = settings.conf['general']['default']
    if name in settings.conf:
        wiki = Wiki.fromName(name)
    else:
        wiki = Wiki.fromApiUrl(name)


def runEditor(text):
    with tempfile.NamedTemporaryFile('w+') as file:
        file.write(text)
        file.flush()
        executeCommand([os.environ.get('EDITOR', 'vim'), file.name])
        file.seek(0)
        return file.read()


def edit(title):
    try:
        text, verify = wiki.init_edit(title)
        wiki.login()

        newtext = runEditor(text)
        if newtext == text:
            ex.notify('Edit Canceled: No Change')
            return

        def submit(button):
            closeOverlay()
            wiki.commit_edit(newtext, summary.edit_text,
                             minor.get_state(), verify)
            openPage(title)
        def cancel(button):
            closeOverlay()
        summary = urwid.Edit('Summary: ')
        minor = urwid.CheckBox('Minor Edit')
        cancel_button = urwid.Button('Cancel', cancel)
        submit_button = urwid.Button('Submit', submit)
        pile = urwid.Pile([summary, minor, cancel_button, submit_button])
        openOverlay(pile, 'Edit', 'pack')
    except WikiError as e:
        ex.notify('Error: ' + str(e))

overlaymap = {'bmarks': Bmarks,
              'wikis': Wikis,
              'feeds': Feeds,
              'contents': Toc,
              'links': Links,
              'iwlinks': Iwlinks,
              'extlinks': Extlinks,
              'langs': Langs}
cmds = tuple(overlaymap) + ('quit', 'bmark', 'open', 'edit', 'clearcache',
                            'help', 'back', 'forward', 'random')

def processCmd(cmd, *args):
    global current

    if cmd in ('q', 'quit'):
        raise urwid.ExitMainLoop
    elif cmd == 'bmark':
        wiki.bmarks.add(page.title)
        ex.notify("Bookmark Added")
    elif cmd in overlaymap:
        openOverlay(overlaymap[cmd]())
    elif cmd == 'open':
        if args:
            openPage(' '.join(args))
        else:
            openOverlay(SearchBox())
    elif cmd == 'clearcache':
        wiki.clear_cache()
    elif cmd == 'edit':
        edit(page.title)
    elif cmd == 'help':
        executeCommand(['man', 'wikicurses'])
    elif cmd == 'back':
        if current > 0:
            current -= 1
            openPage(history[current], browsinghistory=True)
    elif cmd == 'forward':
        if current < len(history)-1:
            current += 1
            openPage(history[current], browsinghistory=True)
    elif cmd == 'random':
        openPage(wiki.random())
    elif cmd:
        ex.notify(cmd + ': Unknown Command')


def openOverlay(widget, title=None, height=('relative', 50), width=('relative', 50)):
    if widget.sizing() == {'flow'}:
        height = 'pack'
    box = urwid.LineBox(widget, title or widget.title)
    overlay = urwid.Overlay(box, mainwidget, 'center', width, 'middle', height)
    loop.widget = overlay


def closeOverlay():
    loop.widget = mainwidget


def main():
    parser = argparse.ArgumentParser(
        description="A simple curses interface for accessing Wikipedia.")

    parser.add_argument('search',
                        nargs='?',
                        help="page to search for")
    parser.add_argument('-w', '--wiki', help='wiki api url')
    # For shell completion functions
    parser.add_argument('--complete',
                        action='store',
                        help=argparse.SUPPRESS)
    parser.add_argument('--dumpcolors',
                        action='store_true',
                        help="print default color settings")

    parser.add_argument('-f', '--feed', help='view featured feed')

    args = parser.parse_args()
    openWiki(args.wiki)

    if args.complete:
        if args.complete == 'search':
            sugestions = wiki.search_sugestions(args.search)
        elif args.complete == 'feed':
            sugestions = wiki.list_featured_feeds()
        elif args.complete == 'wiki':
            sugestions = settings.wikis().keys()
        print(*sugestions, sep='\n')
        return
    elif args.dumpcolors:
        settings.dumpColors()
        print("Color settings written to " + settings.colorspath + '.')
        return

    callback = lambda x, y: openPage(args.feed or args.search, bool(args.feed))
    loop.set_alarm_in(.001, callback)  # Open page once loop is started
    try:
        loop.run()
    except KeyboardInterrupt:
        pass


history = []
current = -1
page = None

palette = []
for x in range(1, sum(formats) + 1):
    fgcolor = settings.getColor(x, 'fgcolor')
    bg = settings.getColor(x, 'bgcolor')
    fgfmts = {j for i in formats if x & i and i.name in settings.colors
            for j in settings.colors[i.name][0]}
    if fgcolor:
        fgfmts.add(fgcolor)
    fg = ','.join(fgfmts)
    palette.append((x, fg, bg))

urwid.command_map['k'] = 'cursor up'
urwid.command_map['j'] = 'cursor down'
urwid.command_map['ctrl b'] = 'cursor page up'
urwid.command_map['ctrl f'] = 'cursor page down'

ex = Ex()
header = urwid.Text('Wikicurses', align='center')
loading = urwid.Filler(urwid.Text('Loading...'), 'top')
progress = urwid.Text('')
footer = urwid.Columns([ex, ('pack', progress)], 2)
mainwidget = urwid.Frame(loading, urwid.AttrMap(header, formats.h1), footer)
loop = urwid.MainLoop(mainwidget, palette=palette, handle_mouse=settings.mouse)



import re
from collections import OrderedDict

import bs4

from wikicurses import formats
from wikicurses import settings

skipclass = ('wiki-sidebar', 'infobox', 'mw-editsection', 'editsection',
             'wikitable', 'thumb', 'gallery', 'article-thumb', 'infobox_v2',
             'mw-headline-anchor', 'toc', 'noprint', 'wikia-gallery')
skipsection = ('External links', 'See also')

if settings.hide_references:
    skipclass += ('reference',)
    skipsection += ('References',)


def parseArticle(html):
    """Parse article html and return list of  (format, text) tuples."""
    html = html.replace('\t', ' ')
    soup = bs4.BeautifulSoup(html, 'lxml')
    # Turn into tuple since the iterator is being modified
    for i in tuple(soup.strings):
        if not {'pre', 'code'}.intersection(j.name for j in i.parents):
            i.replace_with(i.replace('\n', ''))
    for i in soup.find_all('p'):
        i.insert_after(soup.new_string('\n\n'))
    for i in soup.find_all(['h2', 'h3', 'h4', 'h5', 'h6', 'br', 'pre']):
        i.insert_after(soup.new_string('\n'))
    for i in soup.find_all('div'):
        if i.text and not i.find_all('div'):
            i.insert_after(soup.new_string('\n'))
    for i in soup.find_all('li'):
        if i.parent.name == 'ol':
            num = i.parent.find_all('li').index(i) + 1
            i.insert_before(soup.new_string(str(num) + '. '))
        else:
            i.insert_before(soup.new_string('• '))
        i.insert_after(soup.new_string('\n'))
        if 'li' not in (j.name for j in i.next_siblings): # Last item in list
            i.insert_after(soup.new_string('\n'))

    for i in soup.find_all(True, class_=skipclass):
        i.decompose()
    for i in soup.find_all('script'):
        i.decompose()
    for i in soup.find_all('h2'):
        if i.text in skipsection:
            for j in tuple(i.next_elements):
                # Test it is not a str to avoid using str.find() by accident
                if not isinstance(j, str) and (j.name == 'h2' or j.find('h2')):
                    break
                j.extract()
            i.decompose()

    items = []
    for item in soup.strings:
        if not item:
            continue

        if isinstance(item, bs4.element.Comment):
            continue # Strip out html comments

        partags = {i.name for i in item.parents}
        tformat = sum(
            formats[i] for i in set(i.name for i in formats).intersection(partags))
        if [i for i in partags if re.fullmatch('h[3-6]', i)]:
            tformat |= formats.h
        # Make <strong>/<em> use same formats as <b>/<i>
        if 'strong' in partags:
            tformat |= formats.b
        if 'em' in partags:
            tformat |= formats.i

        # Handle divs with padding or borders defined in css style
        styledivs = item.findParents('div', style=True)
        stylekeys = [i.split(':', 1)[0].strip()
                for div in styledivs for i in div.get('style').split(';')]
        if 'padding' in stylekeys:
            tformat |= formats.divpadding
        if 'border' in stylekeys:
            tformat |= formats.divborder

        # Added specifically for handling spaces between removed references
        if items and items[-1][1] and (items[-1][1][-1] == item[0] == ' '):
            item = item[1:]

        # If format same as previous, combine
        if items and items[-1][0] == tformat:
            items[-1][1] += str(item)
        else:
            items.append([tformat, str(item)])

    return items


def parseFeature(html):
    """Parse featured feed html by striping out html tags."""
    # TODO: Support html tags like <b>
    text = bs4.BeautifulSoup(html, 'lxml').text.strip() + '\n\n'
    text = re.sub('\n\n+', '\n\n', text)
    return text


def _processDisambigSection(section):
    items = []
    for i in section:
        if isinstance(i, str):
            continue
        if i.name == 'h2' or i.find('h2'):
            break
        for item in i.find_all('li'):
            items.append(
                (item.a.text if item.a else '', item.text.split('\n')[0]))
    return items


def parseDisambig(html):
    """Parse disambiguation page and return list of (article, text) tuples."""
    sections = OrderedDict()
    soup = bs4.BeautifulSoup(html, 'lxml')
    for i in soup.find_all(True, class_=skipclass):
        i.decompose()
    sections[''] = _processDisambigSection(soup)
    for i in soup.find_all('h2'):
        if i.text not in ('Contents', 'See also'):
            sections[i.text] = _processDisambigSection(i.next_siblings)
    return sections



import os
import json
import collections
import configparser
from wikicurses import formats
from urllib.parse import urlparse

default_configdir = os.environ['HOME'] + '/.config'
configpath = os.environ.get(
    'XDG_CONFIG_HOME', default_configdir) + '/wikicurses'

conf = configparser.ConfigParser()
conf.read(['/etc/wikicurses.conf', configpath + '/config'])

try:
    mouse = conf.getboolean('general', 'mouse')
except (ValueError, configparser.NoOptionError, configparser.NoSectionError):
    mouse = False
try:
    hide_references = conf.getboolean('general', 'hide_references')
except (ValueError, configparser.NoOptionError, configparser.NoSectionError):
    hide_references = False

Attribute = collections.namedtuple('Attribute',
        ('settings', 'fgcolor', 'bgcolor', 'align', 'padding', 'border'))
defcolors = {
        'b': Attribute(['bold',], '', '', '', 0, False),
        'blockquote': Attribute([], 'dark gray', '', '', 0, False),
        'searchresult': Attribute(['standout'], '', '', '', 0, False),
        'h1': Attribute(['bold'], '', 'dark blue', '', 0, False),
        'h2': Attribute(['bold', 'underline'], 'dark blue', '', 'center', 0, False),
        'h': Attribute(['bold', 'underline'], 'dark blue', '', '', 0, False),
        'pre': Attribute([], 'dark green', '', '', 3, False),
        'code': Attribute([], '', '', '', 0, False),
        'divpadding': Attribute([], '', '', '', 6, False),
        'divborder': Attribute([], '', '', '', 0, True),
        }
colorspath = configpath + '/colors'
if os.path.exists(colorspath):
    colorsconf = configparser.ConfigParser()
    colors = {}
    colorsconf.read(colorspath)
    for name, defattr in defcolors.items():
        try:
            settings = colorsconf.get(name, 'settings').split()
        except (configparser.NoSectionError, configparser.NoOptionError):
            settings = defattr.settings
        fgcolor = colorsconf.get(name, 'fgcolor', fallback=defattr.fgcolor)
        bgcolor = colorsconf.get(name, 'bgcolor', fallback=defattr.bgcolor)
        align = colorsconf.get(name, 'align', fallback=defattr.align)
        padding = colorsconf.getint(name, 'padding', fallback=defattr.padding)
        border = colorsconf.getboolean(name, 'border', fallback=defattr.border)
        colors[name] = Attribute(settings,
                fgcolor, bgcolor, align, padding, border)
else:
    colors = defcolors

def getColor(tformat, name, default=''):
    values = [j for j in (getattr(colors[i.name], name) for i in formats
              if tformat & i and i.name in colors) if j]
    return values[-1] if values else default

def dumpColors():
    colorsconf = configparser.ConfigParser()
    for name, (settings, fgcolor, bgcolor, align, padding, border) in colors.items():
        colorsconf.add_section(name)
        colorsconf.set(name, 'settings', ' '.join(settings))
        colorsconf.set(name, 'fgcolor', fgcolor)
        colorsconf.set(name, 'bgcolor', bgcolor)
        colorsconf.set(name, 'align', align)
        colorsconf.set(name, 'padding', str(padding))
        colorsconf.set(name, 'border', str(border))
    with open(colorspath, 'w') as file:
        colorsconf.write(file)

class Settings:

    def __init__(self, wiki, name):
        self.configpath = configpath + '/' + urlparse(wiki).netloc
        self.file = self.configpath + '/' + name

    def __iter__(self):
        if not os.path.exists(self.file):
            return iter(())
        with open(self.file) as file:
            yield from json.load(file)

    def _save(self, bookmarks):
        if not os.path.exists(self.configpath):
            os.makedirs(self.configpath)
        with open(self.file, 'w') as file:
            json.dump(bookmarks, file)

    def add(self, bmark):
        bookmarks = set(self)
        bookmarks.add(bmark)
        self._save(list(bookmarks))

    def discard(self, bmark):
        bookmarks = set(self)
        bookmarks.discard(bmark)
        self._save(list(bookmarks))


def wikis():
    """Return dictionary of wiki urls by name."""
    exclude = ('general', 'DEFAULT', 'keymap')
    return {k: v['url'] for k, v in conf.items() if k not in exclude}



import sys
import time
import hashlib
import json
import urllib.request
import http.cookiejar
from functools import lru_cache

from bs4 import BeautifulSoup

from wikicurses.htmlparse import parseArticle, parseFeature
from wikicurses.settings import Settings, wikis, conf
from wikicurses import formats, __version__

useragent = "Wikicurses/%s (https://github.com/ids1024/wikicurses)"\
            " Python-urllib/%d.%d" % (( __version__,) + sys.version_info[:2])


class WikiError(Exception):
    pass


class Wiki(object):
    """A mediawiki wiki."""
    csrftoken = None
    _articlepath = None
    _mainpage = None

    def __init__(self, url, username, password):
        self.siteurl = url
        self.username = username
        self.password = password
        self.bmarks = Settings(url, 'bookmarks')

    @classmethod
    def fromName(cls, name):
        url = conf[name]['url']
        username = conf[name].get('username')
        password = conf[name].get('password')
        return cls(url, username, password)

    @classmethod
    def fromApiUrl(cls, url):
        wikiname = {v: k for k, v in wikis().items()}.get(url)
        username = password = None
        if wikiname:
            username = conf[wikiname].get('username')
            password = conf[wikiname].get('password')
        return cls(url, username, password)

    @classmethod
    def fromPageUrl(cls, url):
        html = urllib.request.urlopen(url).read().decode()
        soup = BeautifulSoup(html, 'lxml')
        link = soup.find('link', rel='EditURI')
        if not link:
            return None
        rsdurl = urllib.parse.urljoin(url, link['href'])
        rsd = urllib.request.urlopen(rsdurl).read().decode()
        soup = BeautifulSoup(rsd, "xml")
        apiurl = soup.find('api', {'name': 'MediaWiki'})['apiLink']
        return cls.fromApiUrl(apiurl)

    def _get_siteinfo(self):
        result = self._query(action="query", meta="siteinfo",
                             siprop="general", format="json")
        query = json.loads(result)["query"]
        self._articlepath = urllib.parse.urljoin(
            query['general']['base'],
            query['general']['articlepath'])
        self._mainpage = query['general']['mainpage']

    @property
    def articlepath(self):
        if not self._articlepath:
            self._get_siteinfo()
        return self._articlepath

    @property
    def mainpage(self):
        if not self._mainpage:
            self._get_siteinfo()
        return self._mainpage

    def _query(self, post=False, **kwargs):
        params = {k: v for k, v in kwargs.items() if v is not False}
        data = urllib.parse.urlencode(params)
        url = self.siteurl
        if post:
            data = data.encode()
        if not post:
            url += '?' + data
            data = None
        return urllib.request.urlopen(url, data).read().decode('utf-8')

    def login(self):
        """Log in to wiki using stored credentials."""
        if self.csrftoken:  # Already logged in
            return
        if not (self.username or self.password):
            raise WikiError("Username or password not configured for this wiki")
        query = {'post': True, 'action': 'login', 'format': 'json',
                 'lgname': self.username, 'lgpassword': self.password}
        result = json.loads(self._query(**query))['login']
        if result['result'] == 'NeedToken':
            result = json.loads(
                self._query(lgtoken=result['token'], **query))['login']
        if result['result'] != 'Success':  # Error
            raise WikiError(result['result'])
        self.csrftoken = json.loads(self._query(post=True, action='query',
                                                meta='tokens', format='json')
                                    )['query']['tokens']['csrftoken']

    def logout(self):
        """Log out of wiki."""
        self._query(action='logout', format='json')
        self.csrftoken = None

    def init_edit(self, title):
        """Initialize edit of page.

        Return tuple (text, verify) where text is the text to be modified and
        verify should be passed to commit_edit().
        """
        result = json.loads(self._query(action='query', prop='revisions',
                                        rvprop='timestamp|content',
                                        titles=title, format='json'))['query']
        if "missing" in result:
            raise WikiError("Page Not Found")

        rev = next(iter(result['pages'].values()))['revisions'][0]
        starttime = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
        return rev['*'], (title, rev['timestamp'], starttime)

    def commit_edit(self, text, summary, minor, verify):
        """Commit edit of page.
        
        Required arguments:
        text -- the new text to be saved
        summary -- edit summary
        minor -- boolean value; True if edit is minor
        verify -- the verification tuple returned by init_edit()
        """
        md5sum = hashlib.md5(text.encode()).hexdigest()
        result = json.loads(self._query(post=True, action='edit', text=text,
                                        title=verify[0], basetimestamp=verify[1],
                                        starttimestamp=verify[2], md5=md5sum,
                                        token=self.csrftoken, summary=summary,
                                        minor=minor, format='json'))['edit']
        self.search.cache_clear()
        if result['result'] != 'Success':
            raise WikiError(result['result'])

    @lru_cache(16)
    def search(self, name):
        """Search wiki for article and return _Article object."""
        result = json.loads(self._query(action="parse", page=name,
                                        prop="externallinks|iwlinks|langlinks|"
                                        "links|displaytitle|properties|text",
                                        format="json", redirects=True
                                        )).get('parse', {})
        return _Article(name, result)

    @lru_cache(1)
    def list_featured_feeds(self):
        """Return a list of available featured feeds."""
        result = json.loads(self._query(action="paraminfo",
                                        modules="featuredfeed",
                                        format="json"))["paraminfo"]
        if not result["modules"]:
            return []
        return next(i for i in result["modules"][0]["parameters"]
                    if i["name"] == "feed")["type"]

    @lru_cache(16)
    def get_featured_feed(self, feed):
        result = self._query(action="featuredfeed", feed=feed)
        return _Featured(BeautifulSoup(result, "xml").find("channel"))

    @lru_cache(16)
    def search_sugestions(self, name):
        """Return list of search suggestions for specified string."""
        result = self._query(action="opensearch", search=name, format="json")
        return json.loads(result)[1]

    def random(self):
        """Return the name of a random page."""
        result = json.loads(self._query(action="query", list="random",
            rnnamespace=0, format="json"))
        return result["query"]["random"][0]["title"]

    def clear_cache(self):
        """Clear the cache."""
        self.search.cache_clear()
        self.list_featured_feeds.cache_clear()
        self.get_featured_feed.cache_clear()
        self.search_sugestions.cache_clear()


class _Page:
    properties = {}
    html = ''
    links = []
    iwlinks = []
    extlinks = []
    langlinks = {}


class _Article(_Page):
    content = [(0, 'Page Not Found.')]

    def __init__(self, search, result):
        self.title = result.get('title', search)
        self.exists = result != {}
        if self.exists:
            self.properties = {i['name']: i['*'] for i in result.get('properties',())}
            self.html = result['text']['*']
            self.links = [i['*'] for i in result['links'] if ('exists' in i) and
                          not any(i['*'].startswith(j + ':') for j in
                          ('Category', 'Template', 'Template talk', 'Wikipedia'))]
            self.iwlinks = [(i['*'].split(':', 1)[1], i['url'])
                            for i in result['iwlinks']]
            # if an url starts with //, it can by http or https.  Use http.
            self.extlinks = ['http:' + i if i.startswith('//') else i
                             for i in result['externallinks']]
            self.langlinks = {i.get('autonym', i['lang']): (i['url'], i['*'])
                             for i in result.get('langlinks')}

            self.content = parseArticle(self.html)
            if self.extlinks:
                self.content.append([formats.h2, 'External links'])
                self.content.append([0, '\n'.join(self.extlinks) + '\n'])


class _Featured(_Page):
    exists = True

    def __init__(self, result):
        self.title = result.find('title').text
        self.content = []
        self.content.append([0, parseFeature(result.find('description').text)])
        for i in result.find_all('item'):
            description = i.find('description').text
            text = parseFeature(description)
            self.content.append([formats.h2, i.find('title').text])
            self.content.append([0, text])


cookiejar = http.cookiejar.CookieJar()
opener = urllib.request.build_opener(
    urllib.request.HTTPCookieProcessor(cookiejar))
opener.addheaders = [('User-agent', useragent)]
urllib.request.install_opener(opener)
